#!/bin/sh
#
#  Copyright 2014- Stichting Fundamenteel Onderzoek der Materie (FOM-Nikhef)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# Demonstrator script to convert a PAP policy into a banmapfile (on stdout) for
# use with LCAS and LCMAPS plugins.
# Adapted by Mischa Salle (msalle@nikhef.nl) from a curl one-liner from Vincent
# Brillault.
#
# Example:
# get-argus-denylist.sh argus.testmachine central ngi
# where central and ngi are aliases for remote PAPs defined for the Argus-PAP
# running on argus.testmachine

# debug flag by default is unset
DEBUG=0

capath=/etc/grid-security/certificates
cert=/etc/grid-security/hostcert.pem
key=/etc/grid-security/hostkey.pem
proxy=''
# Flag marking the method of authentication to use
# 1 - host certificate authentication (default)
# 2 - proxy certificate 
AUTH_METHOD=0 

# PAP port and SOAP endpoint
port=8150
endp='/pap/services/XACMLPolicyManagementService?wsdl'

#command line options
VERBOSE=0
outfile=''
rawfile=''

# Temporary files will be based on program name
prog=$(basename $0 .sh)
template=${prog}_XXXXXX

auth_error="Only one authentication method is accepted exclusively! (use either host OR proxy certificate)"

# Error function
error()	{
    echo "$prog: $*" >&2
    exit 1
}

# Called when we receive an empty but non-error result
endstat()	{
    echo "$prog: $*" >&2
    exit 0
}

# Print message to stderr when verbosity is enabled
verbose()	{
    if [ $VERBOSE -ne 0 ] ; then
	echo "$*" >&2
    fi
}

# clean up the temprorary files and reset traps
cleanup()	{
     exitcode=$?
     trap - 0 1 2 3 15
     if [ $DEBUG -eq 0 ] ; then 
     	rm -f $raw_xacml $parsed_xacml $curl_output $deny_list $xmlgrep_error 2> /dev/null
     fi
     if [ $1 -eq 0 ] ; then
        exit $exitcode
     else
     	exit $1
     fi
}

# Print help message with command usage and short explanation of options
help()	{
    echo "Usage: $prog [option] <PAP host> [alias...]"
    echo ''
    echo 'Retrieves a list of policies (XACML) from the Argus PAP banning service for each alias,'
    echo 'and extracts the identities of the subjects of every "Deny" rule.'
    echo ''
    echo 'Options:'
    echo '	-h / --help			Print this message'
    echo '	--key				Key used for authentication'
    echo '	--cert				Certificate used for authentication'
    echo '	--proxy				Use a proxy certificate instead of host certificate'
    echo '	-o FILE / --output FILE		Output list of denied identities to FILE'
    echo '	-r FILE / --raw FILE		Dump raw XACML output to FILE'
    echo '	-v / --verbose			Print verbose messages'
    echo ''
}

# Helper function to convert RFC2253 DNs into old-style OpenSSL syntax
# It splits on comma's, reverses the order and prefixes each part with a slash
convert_dn()	{
    # Only convert when it contains CN=
    if echo "$1"|grep -q "^/";then
	# Looks like an FQAN 
	echo "\"$1\""
    elif echo "$1"|grep -q "CN=";then
	# Looks like a DN
	printf "\""
	echo "$1" | tr , '\n' | tac | while read a ; do printf '%s' "/$a" ; done
	echo "\""
    else
	error "Illegal line in output results: \"$1\""
    fi
}

# echoes the postdata for given alias
get_postdata()	{
    alias=$1
    printf '<?xml version="1.0" encoding="UTF-8"?>'
    printf '<SOAP-ENV:Envelope xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/" xmlns:ns0="http://schemas.xmlsoap.org/soap/encoding/" xmlns:ns1="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns2="http://services.pap.authz.glite.org" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">'
    printf '<SOAP-ENV:Header/>'
    printf '%s' "<ns1:Body><ns2:listPolicies><papAlias>$alias</papAlias></ns2:listPolicies></ns1:Body>"
    printf '</SOAP-ENV:Envelope>\n'
}

# Runs actual curl command to obtain XACML policy
run_curl()  {
    alias=$1
    
    if [ -z "$alias" ] ; then
	error "Internal error, $0 did not receive obligatory argument"
    fi

    optional_args=""
    if [ $VERBOSE -ne 0 ] ; then
	optional_args="$optional_args -v"
    else
	optional_args="$optional_args -s"
    fi
 
    if [ $AUTH_METHOD -eq 1 ] ; then
	optional_args="$optional_args --cert $cert --key $key"
    elif [ $AUTH_METHOD -eq 2 ] ; then
	optional_args="$optional_args --cert $proxy --key $proxy --cacert $proxy"
    fi
	
   
    # Run curl, insert HTTP return code as comment (such that xml_grep won't
    # mind) at the end.
    curl \
        $optional_args \
	-S \
	-w "\n<!-- http_code=%{http_code} -->" \
	--capath $capath \
	-H 'SOAPAction: ""' \
	-H "Content-Type: text/xml; charset=utf-8" \
	-X POST \
	-d "$(get_postdata $alias)" \
	-o $raw_xacml \
	"https://${host}:${port}${endp}"

    return $?
}

#Set trap for exit signals to call cleanup
for sig in 0 1 2 3 15 ; do
	trap "cleanup $sig" $sig
done

umask 0177

raw_xacml=$(mktemp --tmpdir $template) # response of the curl request (XACML policies)
parsed_xacml=$(mktemp --tmpdir $template) # output of xml_grep for deny rules 
curl_output=$(mktemp --tmpdir $template) # output of curl
deny_list=$(mktemp --tmpdir $template) # output of the script (deny list)
xmlgrep_error=$(mktemp --tmpdir $template) # redirect of stderr on xml_grep execution

if [ $? -ne 0 ] ; then
    error "Failed to create temporary files!"
fi

# Command line [option] validation and setting
options=`getopt -o v,o:,h,r: -l output:,verbose,help,raw:,key:,cert:,proxy: -n "$prog" -- "$@"`

if [ $? -ne 0 ] ; then 
    help
    exit 1
fi

# Replace positional parameter list with validated option -- parameters
eval set -- $options

while [ $1 != '--' ] ; do
	case $1 in
	-v | --verbose)  VERBOSE=1
	    		 ;;
	-o | --output )  outfile=$2
			 shift
	    		 ;;
	-r | --raw )  	 rawfile=$2
			 shift
	    		 ;;
	--key )  	 key=$2
			 if [ $AUTH_METHOD -gt 1 ] ; then 
			 	error "$auth_error" 
			 fi
			 AUTH_METHOD=1
			 shift
	    		 ;;
	--cert )  	 cert=$2
			 if [ $AUTH_METHOD -gt 1 ] ; then 
			 	error "$auth_error" 
			 fi
			 AUTH_METHOD=1
			 shift
	    		 ;;
	--proxy )  	 proxy=$2
			 if [ $AUTH_METHOD -gt 0 ] ; then 
			 	error "$auth_error" 
			 fi
			 AUTH_METHOD=2
			 shift
	    		 ;;
	-h | --help )    help
			 exit 0
			 ;; 
	* )  		 help 
	    		 exit 1
	    		 ;;
	esac
	shift
done

#skip the '--'
shift

# Input Validation
if [ $# -lt 1 ];then
    help
    error "Invalid Usage"
fi
host=$1 ; shift
if [ -n "$1" ];then
    aliases=$@
else
    aliases=default
fi

#set default authentication method in case none of the
# --key --cert or --proxy options have been set
if [ $AUTH_METHOD -gt 1 ] ; then 
    if [ ! -r "$proxy" ] ; then
	error "Cannot read proxy from $key"
    fi
else
    AUTH_METHOD=1
    #verify existence of the required $key and $cert 
    if [ ! -r "$key" ] ; then
	error "Cannot read key from $key"
    fi

    if [ ! -r "$cert" ] ; then
	error "Cannot read certificate from $cert"
    fi
fi

# Start of actual program

# Loop over aliases, unfortunately we cannot get the policies from all remote
# PAPs simultaneously so we need to query them separately. Since the different
# XACML outputs have different xml roots, we also do the xml parsing for each
# individually.
for alias in $aliases ; do
    
    verbose  "Downloading deny list from https://${host}:${port}${endp} with alias $alias..."  

    run_curl $alias  > $curl_output
	
    verbose "curl has finished with exit code $? and response code `grep 'http_code' $curl_output`" 

    # Check curl did not fail and that we got a HTTP 200 code 
    if [ $? != 0 -o -z "$(grep 'http_code=200' $curl_output)" ] ;then   
	error "Cannot download XACML for $alias from $host"
    fi

    # Dump XACML
    if [ ! -z "$rawfile" ] ; then
    	cp -f $raw_xacml "$rawfile"
    fi
  
    verbose "Stripping deny rules..."
 
    # Get the Deny rules
    xml_grep \
	--root 'xacml:Rule[@Effect="Deny"]' $raw_xacml > $parsed_xacml 2> $xmlgrep_error

    # Parse return value
    if [ $? -ne 0 -o -s "$xmlgrep_error"  ];then
	cat $xmlgrep_error
	error "xml_grep for Deny rules failed"
    fi

    # Check we got Deny rules
    if [ -s "$parsed_xacml" ];then
	# Look for subject-id and fqans
	xml_grep \
	    --root 'xacml:SubjectMatch' \
	    --cond '[@AttributeId="http://glite.org/xacml/attribute/fqan"]' \
	    --cond '[@AttributeId="urn:oasis:names:tc:xacml:1.0:subject:subject-id"]' \
	    --pretty_print indented \
	    --text_only \
	    $parsed_xacml |\
	    while read line;do
		# convert RFC2253 format into OpenSSL format
		convert_dn "$line" >> $deny_list
	    done
    fi
done

# Dump the output
if [ -z "$outfile" ] ; then
    cat $deny_list 
else
    verbose  "Dumping output into $outfile..."
    mv -f $deny_list "$outfile"
fi
